@@ -1,6 +1,6 @@
 /*
  * ProFTPD - mod_sftp 'password' user authentication
- * Copyright (c) 2008-2014 TJ Saunders
+ * Copyright (c) 2008-2015 TJ Saunders
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -20,8 +20,6 @@
  * give permission to link this program with OpenSSL, and distribute the
  * resulting executable, without including the source code for OpenSSL in the
  * source distribution.
- *
- * $Id: auth-password.c,v 1.10 2014/03/02 22:05:43 castaglia Exp $
  */
 
 #include "mod_sftp.h"
@@ -57,6 +55,12 @@ int sftp_auth_password(struct ssh2_packet *pkt, cmd_rec *pass_cmd,
         "cipher algorithm '%s' or MAC algorithm '%s' unacceptable for "
         "password authentication, denying password authentication request",
         cipher_algo, mac_algo);
+
+      pr_log_auth(PR_LOG_NOTICE,
+        "USER %s (Login failed): cipher algorithm '%s' or MAC algorithm '%s' "
+        "unsupported for password authentication", user,
+        cipher_algo, mac_algo);
+
       *send_userauth_fail = TRUE;
       errno = EPERM;
       return 0;
@@ -79,7 +83,11 @@ int sftp_auth_password(struct ssh2_packet *pkt, cmd_rec *pass_cmd,
   if (pr_cmd_dispatch_phase(pass_cmd, PRE_CMD, 0) < 0) {
     (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
       "authentication request for user '%s' blocked by '%s' handler",
-      orig_user, pass_cmd->argv[0]);
+      orig_user, (char *) pass_cmd->argv[0]);
+
+    pr_log_auth(PR_LOG_NOTICE,
+      "USER %s (Login failed): blocked by '%s' handler", orig_user,
+      (char *) pass_cmd->argv[0]);
 
     pr_cmd_dispatch_phase(pass_cmd, POST_CMD_ERR, 0);
     pr_cmd_dispatch_phase(pass_cmd, LOG_CMD_ERR, 0);
@@ -164,3 +172,7 @@ int sftp_auth_password(struct ssh2_packet *pkt, cmd_rec *pass_cmd,
 
   return 1;
 }
+
+int sftp_auth_password_init(pool *p) {
+  return 0;
+}
