@@ -1,6 +1,6 @@
 /*
  * ProFTPD - mod_sftp keyboard-interactive driver mgmt
- * Copyright (c) 2008-2015 TJ Saunders
+ * Copyright (c) 2008-2017 TJ Saunders
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -31,6 +31,8 @@
 
 #define SFTP_KBDINT_MAX_RESPONSES	500
 
+extern pr_response_t *resp_list, *resp_err_list;
+
 struct kbdint_driver {
   struct kbdint_driver *next, *prev;
 
@@ -190,7 +192,7 @@ int sftp_kbdint_unregister_driver(const char *name) {
  */
 
 int sftp_kbdint_send_challenge(const char *user, const char *instruction,
-    unsigned int count, sftp_kbdint_challenge_t *challenges) {
+    uint32_t count, sftp_kbdint_challenge_t *challenges) {
   register unsigned int i;
   unsigned char *buf, *ptr;
   uint32_t buflen, bufsz;
@@ -252,8 +254,8 @@ int sftp_kbdint_send_challenge(const char *user, const char *instruction,
   return res;
 }
 
-int sftp_kbdint_recv_response(pool *p, unsigned int expected_count,
-    unsigned int *rcvd_count, const char ***responses) {
+int sftp_kbdint_recv_response(pool *p, uint32_t expected_count,
+    uint32_t *rcvd_count, const char ***responses) {
   register unsigned int i;
   unsigned char *buf;
   cmd_rec *cmd;
@@ -262,6 +264,7 @@ int sftp_kbdint_recv_response(pool *p, unsigned int expected_count,
   struct ssh2_packet *pkt;
   char mesg_type;
   int res;
+  pool *resp_pool = NULL;
 
   if (p == NULL ||
       rcvd_count == NULL ||
@@ -278,12 +281,20 @@ int sftp_kbdint_recv_response(pool *p, unsigned int expected_count,
     return res;
   }
 
+  pr_response_clear(&resp_list);
+  pr_response_clear(&resp_err_list);
+
+  /* Cache a reference to the current response pool used. */
+  resp_pool = pr_response_get_pool();
+  pr_response_set_pool(pkt->pool);
+
   mesg_type = sftp_ssh2_packet_get_mesg_type(pkt);
   if (mesg_type != SFTP_SSH2_MSG_USER_AUTH_INFO_RESP) {
     (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
       "expecting USER_AUTH_INFO_RESP message, received %s (%d)",
       sftp_ssh2_packet_get_mesg_type_desc(mesg_type), mesg_type);
     destroy_pool(pkt->pool);
+    pr_response_set_pool(resp_pool);
     errno = EPERM;
     return -1;
   }
@@ -309,6 +320,7 @@ int sftp_kbdint_recv_response(pool *p, unsigned int expected_count,
       expected_count != 1 ? "challenges" : "challenge",
       (unsigned long) resp_count, resp_count != 1 ? "responses" : "response");
     destroy_pool(pkt->pool);
+    pr_response_set_pool(resp_pool);
     errno = EPERM;
     return -1;
   }
@@ -318,6 +330,7 @@ int sftp_kbdint_recv_response(pool *p, unsigned int expected_count,
       "received too many responses (%lu > max %lu), rejecting",
       (unsigned long) resp_count, (unsigned long) SFTP_KBDINT_MAX_RESPONSES);
     destroy_pool(pkt->pool);
+    pr_response_set_pool(resp_pool);
     errno = EPERM;
     return -1;
   }
@@ -330,9 +343,10 @@ int sftp_kbdint_recv_response(pool *p, unsigned int expected_count,
     *((char **) push_array(list)) = pstrdup(p, sftp_utf8_decode_str(p, resp));
   }
 
-  *rcvd_count = (unsigned int) resp_count;
+  *rcvd_count = resp_count;
   *responses = ((const char **) list->elts);
   destroy_pool(pkt->pool);
+  pr_response_set_pool(resp_pool);
 
   return 0;
 }
