@@ -1,6 +1,6 @@
 /*
  * ProFTPD - FTP server daemon
- * Copyright (c) 2001-2011 The ProFTPD Project team
+ * Copyright (c) 2001-2016 The ProFTPD Project team
  *  
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -22,9 +22,7 @@
  * OpenSSL in the source distribution.
  */
 
-/* Controls API routines
- * $Id: ctrls.c,v 1.30 2011/06/05 22:45:14 castaglia Exp $
- */
+/* Controls API routines */
 
 #include "conf.h"
 #include "privs.h"
@@ -45,6 +43,15 @@
 
 #include "mod_ctrls.h"
 
+/* Maximum number of request arguments. */
+#define CTRLS_MAX_NREQARGS	32
+
+/* Maximum number of response arguments. */
+#define CTRLS_MAX_NRESPARGS	1024
+
+/* Maximum length of a single request argument. */
+#define CTRLS_MAX_REQARGLEN	256
+
 typedef struct ctrls_act_obj {
   struct ctrls_act_obj *prev, *next;
   pool *pool;
@@ -198,8 +205,9 @@ static char *ctrls_sep(char **str) {
   if (!str || !*str || !**str)
     return NULL;
 
-  while (**str && isspace((int) **str))
+  while (**str && PR_ISSPACE(**str)) {
     (*str)++;
+  }
 
   if (!**str)
     return NULL;
@@ -212,7 +220,7 @@ static char *ctrls_sep(char **str) {
   }
 
   while (**str &&
-         (quoted ? (**str != '\"') : !isspace((int) **str))) {
+         (quoted ? (**str != '\"') : !PR_ISSPACE(**str))) {
 
     if (**str == '\\' && quoted) {
 
@@ -235,10 +243,12 @@ static char *ctrls_sep(char **str) {
 int pr_ctrls_register(const module *mod, const char *action,
     const char *desc, int (*cb)(pr_ctrls_t *, int, char **)) {
   ctrls_action_t *act = NULL, *acti = NULL;
-  int act_id = -1;
+  unsigned int act_id = 0;
 
   /* sanity checks */
-  if (!action || !desc || !cb) {
+  if (action == NULL ||
+      desc == NULL ||
+      cb == NULL) {
     errno = EINVAL;
     return -1;
   }
@@ -334,28 +344,39 @@ int pr_ctrls_unregister(module *mod, const char *action) {
   return 0;
 }
 
-int pr_ctrls_add_arg(pr_ctrls_t *ctrl, char *ctrls_arg) {
+int pr_ctrls_add_arg(pr_ctrls_t *ctrl, char *ctrls_arg, size_t ctrls_arglen) {
+  register unsigned int i;
 
   /* Sanity checks */
-  if (!ctrl || !ctrls_arg) {
+  if (ctrl == NULL ||
+      ctrls_arg == NULL) {
     errno = EINVAL;
     return -1;
   }
 
+  /* Scan for non-printable characters. */
+  for (i = 0; i < ctrls_arglen; i++) {
+    if (!PR_ISPRINT((int) ctrls_arg[i])) {
+      errno = EPERM;
+      return -1;
+    }
+  }
+
   /* Make sure the pr_ctrls_t has a temporary pool, from which the args will
    * be allocated.
    */
-  if (!ctrl->ctrls_tmp_pool) {
+  if (ctrl->ctrls_tmp_pool == NULL) {
     ctrl->ctrls_tmp_pool = make_sub_pool(ctrls_pool);
     pr_pool_tag(ctrl->ctrls_tmp_pool, "ctrls tmp pool");
   }
 
-  if (!ctrl->ctrls_cb_args)
+  if (ctrl->ctrls_cb_args == NULL) {
     ctrl->ctrls_cb_args = make_array(ctrl->ctrls_tmp_pool, 0, sizeof(char *));
+  }
 
   /* Add the given argument */
-  *((char **) push_array(ctrl->ctrls_cb_args)) = pstrdup(ctrl->ctrls_tmp_pool,
-    ctrls_arg);
+  *((char **) push_array(ctrl->ctrls_cb_args)) = pstrndup(ctrl->ctrls_tmp_pool,
+    ctrls_arg, ctrls_arglen);
 
   return 0;
 }
@@ -363,7 +384,8 @@ int pr_ctrls_add_arg(pr_ctrls_t *ctrl, char *ctrls_arg) {
 int pr_ctrls_copy_args(pr_ctrls_t *src_ctrl, pr_ctrls_t *dst_ctrl) {
 
   /* Sanity checks */
-  if (!src_ctrl || !dst_ctrl) {
+  if (src_ctrl == NULL ||
+      dst_ctrl == NULL) {
     errno = EINVAL;
     return -1;
   }
@@ -371,13 +393,14 @@ int pr_ctrls_copy_args(pr_ctrls_t *src_ctrl, pr_ctrls_t *dst_ctrl) {
   /* If source ctrl has no ctrls_cb_args member, there's nothing to be
    * done.
    */
-  if (!src_ctrl->ctrls_cb_args)
+  if (src_ctrl->ctrls_cb_args == NULL) {
     return 0;
+  }
 
   /* Make sure the pr_ctrls_t has a temporary pool, from which the args will
    * be allocated.
    */
-  if (!dst_ctrl->ctrls_tmp_pool) {
+  if (dst_ctrl->ctrls_tmp_pool == NULL) {
     dst_ctrl->ctrls_tmp_pool = make_sub_pool(ctrls_pool);
     pr_pool_tag(dst_ctrl->ctrls_tmp_pool, "ctrls tmp pool");
   }
@@ -502,13 +525,13 @@ int pr_ctrls_parse_msg(pool *msg_pool, char *msg, unsigned int *msgargc,
 
 int pr_ctrls_recv_request(pr_ctrls_cl_t *cl) {
   pr_ctrls_t *ctrl = NULL, *next_ctrl = NULL;
-  char reqaction[512] = {'\0'}, *reqarg = NULL;
+  char reqaction[128] = {'\0'}, *reqarg = NULL;
   size_t reqargsz = 0;
   unsigned int nreqargs = 0, reqarglen = 0;
-  int status = 0;
-  register int i = 0;
+  int bread, status = 0;
+  register unsigned int i = 0;
 
-  if (!cl) {
+  if (cl == NULL) {
     errno = EINVAL;
     return -1;
   }
@@ -527,14 +550,53 @@ int pr_ctrls_recv_request(pr_ctrls_cl_t *cl) {
    * the same function, pr_ctrls_send_msg(), is used to send requests
    * as well as responses, and the status is a necessary part of a response.
    */
-  if (read(cl->cl_fd, &status, sizeof(int)) < 0) {
+  bread = read(cl->cl_fd, &status, sizeof(int));
+  if (bread < 0) {
+    int xerrno = errno;
+
     pr_signals_unblock();
+
+    errno = xerrno;
+    return -1;
+  }
+
+  /* Watch for short reads. */
+  if (bread != sizeof(int)) {
+    (void) pr_trace_msg(trace_channel, 3,
+      "short read (%d of %u bytes) of status, unable to receive request",
+      bread, (unsigned int) sizeof(int));
+    pr_signals_unblock();
+    errno = EPERM;
     return -1;
   }
  
   /* Read in the args, length first, then string. */
-  if (read(cl->cl_fd, &nreqargs, sizeof(unsigned int)) < 0) {
+  bread = read(cl->cl_fd, &nreqargs, sizeof(unsigned int));
+  if (bread < 0) {
+    int xerrno = errno;
+
     pr_signals_unblock();
+
+    errno = xerrno;
+    return -1;
+  }
+
+  /* Watch for short reads. */
+  if (bread != sizeof(unsigned int)) {
+    (void) pr_trace_msg(trace_channel, 3,
+      "short read (%d of %u bytes) of nreqargs, unable to receive request",
+      bread, (unsigned int) sizeof(unsigned int));
+    pr_signals_unblock();
+    errno = EPERM;
+    return -1;
+  }
+
+  if (nreqargs > CTRLS_MAX_NREQARGS) {
+    (void) pr_trace_msg(trace_channel, 3,
+      "nreqargs (%u) exceeds max (%u), rejecting", nreqargs,
+      CTRLS_MAX_NREQARGS);
+    pr_signals_unblock();
+    errno = ENOMEM;
     return -1;
   }
 
@@ -544,8 +606,23 @@ int pr_ctrls_recv_request(pr_ctrls_cl_t *cl) {
    * matching pr_ctrls_t (if present), and add the remaining arguments to it.
    */
   
-  if (read(cl->cl_fd, &reqarglen, sizeof(unsigned int)) < 0) {
+  bread = read(cl->cl_fd, &reqarglen, sizeof(unsigned int));
+  if (bread < 0) {
+    int xerrno = errno;
+
     pr_signals_unblock();
+
+    errno = xerrno;
+    return -1;
+  }
+
+  /* Watch for short reads. */
+  if (bread != sizeof(unsigned int)) {
+    (void) pr_trace_msg(trace_channel, 3,
+      "short read (%d of %u bytes) of reqarglen, unable to receive request",
+      bread, (unsigned int) sizeof(unsigned int));
+    pr_signals_unblock();
+    errno = EPERM;
     return -1;
   }
 
@@ -557,8 +634,23 @@ int pr_ctrls_recv_request(pr_ctrls_cl_t *cl) {
 
   memset(reqaction, '\0', sizeof(reqaction));
 
-  if (read(cl->cl_fd, reqaction, reqarglen) < 0) {
+  bread = read(cl->cl_fd, reqaction, reqarglen);
+  if (bread < 0) {
+    int xerrno = errno;
+
     pr_signals_unblock();
+
+    errno = xerrno;
+    return -1;
+  }
+
+  /* Watch for short reads. */
+  if ((size_t) bread != reqarglen) {
+    (void) pr_trace_msg(trace_channel, 3,
+      "short read (%d of %u bytes) of reqaction, unable to receive request",
+      bread, reqarglen);
+    pr_signals_unblock();
+    errno = EPERM;
     return -1;
   }
 
@@ -571,6 +663,8 @@ int pr_ctrls_recv_request(pr_ctrls_cl_t *cl) {
   ctrl = ctrls_lookup_action(NULL, reqaction, TRUE);
   if (ctrl == NULL) {
     pr_signals_unblock();
+
+    /* XXX This is where we could also add "did you mean" functionality. */
     errno = EINVAL;
     return -1;
   }
@@ -578,8 +672,35 @@ int pr_ctrls_recv_request(pr_ctrls_cl_t *cl) {
   for (i = 0; i < nreqargs; i++) {
     memset(reqarg, '\0', reqargsz);
 
-    if (read(cl->cl_fd, &reqarglen, sizeof(unsigned int)) < 0) {
+    bread = read(cl->cl_fd, &reqarglen, sizeof(unsigned int));
+    if (bread < 0) {
+      int xerrno = errno;
+
       pr_signals_unblock();
+
+      errno = xerrno;
+      return -1;
+    }
+
+    /* Watch for short reads. */
+    if (bread != sizeof(unsigned int)) {
+      (void) pr_trace_msg(trace_channel, 3,
+        "short read (%d of %u bytes) of reqarglen (#%u), skipping",
+        bread, i+1, (unsigned int) sizeof(unsigned int));
+      continue;
+    }
+
+    if (reqarglen == 0) {
+      /* Skip any zero-length arguments. */
+      continue;
+    }
+
+    if (reqarglen > CTRLS_MAX_REQARGLEN) {
+      (void) pr_trace_msg(trace_channel, 3,
+        "reqarglen (#%u) of %u bytes exceeds max (%u bytes), rejecting",
+        i+1, reqarglen, CTRLS_MAX_REQARGLEN);
+      pr_signals_unblock();
+      errno = ENOMEM;
       return -1;
     }
 
@@ -598,13 +719,30 @@ int pr_ctrls_recv_request(pr_ctrls_cl_t *cl) {
       reqarg = pcalloc(ctrl->ctrls_tmp_pool, reqargsz);
     }
 
-    if (read(cl->cl_fd, reqarg, reqarglen) < 0) {
+    bread = read(cl->cl_fd, reqarg, reqarglen);
+    if (bread < 0) {
+      int xerrno = errno;
+
       pr_signals_unblock();
+
+      errno = xerrno;
       return -1;
     }
 
-    if (pr_ctrls_add_arg(ctrl, reqarg)) {
+    /* Watch for short reads. */
+    if ((size_t) bread != reqarglen) {
+      (void) pr_trace_msg(trace_channel, 3,
+        "short read (%d of %u bytes) of reqarg (#%u), skipping",
+        bread, i+1, reqarglen);
+      continue;
+    }
+
+    if (pr_ctrls_add_arg(ctrl, reqarg, reqarglen)) {
+      int xerrno = errno;
+
       pr_signals_unblock();
+
+      errno = xerrno;
       return -1;
     }
   }
@@ -621,9 +759,15 @@ int pr_ctrls_recv_request(pr_ctrls_cl_t *cl) {
    */
   next_ctrl = ctrls_lookup_next_action(NULL, TRUE);
 
-  while (next_ctrl) {
-    if (pr_ctrls_copy_args(ctrl, next_ctrl))
+  while (next_ctrl != NULL) {
+    if (pr_ctrls_copy_args(ctrl, next_ctrl) < 0) {
+      int xerrno = errno;
+
+      pr_signals_unblock();
+
+      errno = xerrno;
       return -1;
+    }
 
     /* Add this ctrl object to the client object. */
     *((pr_ctrls_t **) push_array(cl->cl_ctrls)) = next_ctrl;
@@ -641,13 +785,15 @@ int pr_ctrls_recv_request(pr_ctrls_cl_t *cl) {
 
 int pr_ctrls_recv_response(pool *resp_pool, int ctrls_sockfd,
     int *status, char ***respargv) {
-  register int i = 0;
+  register unsigned int i = 0;
   array_header *resparr = NULL;
   unsigned int respargc = 0, resparglen = 0;
   char response[PR_TUNABLE_BUFFER_SIZE] = {'\0'};
 
   /* Sanity checks */
-  if (!resp_pool || ctrls_sockfd < 0 || !status) {
+  if (resp_pool == NULL ||
+      ctrls_sockfd < 0 ||
+      status == NULL) {
     errno = EINVAL;
     return -1;
   }
@@ -671,6 +817,15 @@ int pr_ctrls_recv_response(pool *resp_pool, int ctrls_sockfd,
     return -1;
   }
 
+  if (respargc > CTRLS_MAX_NRESPARGS) {
+    (void) pr_trace_msg(trace_channel, 3,
+      "respargc (%u) exceeds max (%u), rejecting", respargc,
+      CTRLS_MAX_NRESPARGS);
+    pr_signals_unblock();
+    errno = ENOMEM;
+    return -1;
+  }
+
   /* Read each response, and add it to the array */ 
   for (i = 0; i < respargc; i++) {
     int bread = 0, blen = 0;
@@ -691,7 +846,7 @@ int pr_ctrls_recv_response(pool *resp_pool, int ctrls_sockfd,
     memset(response, '\0', sizeof(response));
 
     bread = read(ctrls_sockfd, response, resparglen);
-    while (bread != resparglen) {
+    while ((size_t) bread != resparglen) {
       if (bread < 0) {
         pr_signals_unblock(); 
         return -1;
@@ -707,8 +862,9 @@ int pr_ctrls_recv_response(pool *resp_pool, int ctrls_sockfd,
     *((char **) push_array(resparr)) = pstrdup(resp_pool, response);
   }
 
-  if (respargv)
+  if (respargv) {
     *respargv = ((char **) resparr->elts);
+  }
 
   pr_signals_unblock(); 
   return respargc;
@@ -716,7 +872,7 @@ int pr_ctrls_recv_response(pool *resp_pool, int ctrls_sockfd,
 
 int pr_ctrls_send_msg(int sockfd, int msgstatus, unsigned int msgargc,
     char **msgargv) {
-  register int i = 0;
+  register unsigned int i = 0;
   unsigned int msgarglen = 0;
 
   /* Sanity checks */
@@ -725,11 +881,13 @@ int pr_ctrls_send_msg(int sockfd, int msgstatus, unsigned int msgargc,
     return -1;
   }
 
-  if (msgargc < 1)
-    return 0;    
-
-  if (msgargv == NULL)
+  if (msgargc < 1) {
     return 0;
+  }
+
+  if (msgargv == NULL) {
+    return 0;
+  }
 
   /* No interruptions */
   pr_signals_block();
@@ -759,28 +917,30 @@ int pr_ctrls_send_msg(int sockfd, int msgstatus, unsigned int msgargc,
       res = write(sockfd, &msgarglen, sizeof(unsigned int));
 
       if (res != sizeof(unsigned int)) {
-        if (errno == EAGAIN)
+        if (errno == EAGAIN) {
           continue;
+        }
 
         pr_signals_unblock();
         return -1;
+      }
 
-      } else
-        break;
+      break;
     }
 
     while (TRUE) {
       res = write(sockfd, msgargv[i], msgarglen);
-
-      if (res != msgarglen) {
-        if (errno == EAGAIN)
+      if ((size_t) res != msgarglen) {
+        if (errno == EAGAIN) {
           continue;
+        }
 
         pr_signals_unblock();
         return -1;
 
-      } else
-        break;
+      }
+
+      break;
     }
   }
 
@@ -942,7 +1102,21 @@ int pr_ctrls_connect(const char *socket_file) {
   /* Create a Unix domain socket */
   sockfd = socket(AF_UNIX, SOCK_STREAM, 0);
   if (sockfd < 0) {
+    int xerrno = errno;
+
     pr_signals_unblock();
+
+    errno = xerrno;
+    return -1;
+  }
+
+  if (fcntl(sockfd, F_SETFD, FD_CLOEXEC) < 0) {
+    int xerrno = errno;
+
+    (void) close(sockfd);
+    pr_signals_unblock();
+
+    errno = xerrno;
     return -1;
   }
 
@@ -962,19 +1136,29 @@ int pr_ctrls_connect(const char *socket_file) {
   len = sizeof(cl_sock);
 
   /* Make sure the file doesn't already exist */
-  unlink(cl_sock.sun_path);
+  (void) unlink(cl_sock.sun_path);
 
   /* Make it a socket */
   if (bind(sockfd, (struct sockaddr *) &cl_sock, len) < 0) {
-    unlink(cl_sock.sun_path);
+    int xerrno = errno;
+
+    (void) unlink(cl_sock.sun_path);
+    (void) close(sockfd);
     pr_signals_unblock();
+
+    errno = xerrno;
     return -1;
   }
 
   /* Set the proper mode */
   if (chmod(cl_sock.sun_path, PR_CTRLS_CL_MODE) < 0) {
-    unlink(cl_sock.sun_path);
+    int xerrno = errno;
+
+    (void) unlink(cl_sock.sun_path);
+    (void) close(sockfd);
     pr_signals_unblock();
+
+    errno = xerrno;
     return -1;
   }
 
@@ -982,20 +1166,30 @@ int pr_ctrls_connect(const char *socket_file) {
   memset(&ctrl_sock, 0, sizeof(ctrl_sock));
 
   ctrl_sock.sun_family = AF_UNIX;
-  strncpy(ctrl_sock.sun_path, socket_file, strlen(socket_file));
+  sstrncpy(ctrl_sock.sun_path, socket_file, sizeof(ctrl_sock.sun_path));
   len = sizeof(ctrl_sock);
 
   if (connect(sockfd, (struct sockaddr *) &ctrl_sock, len) < 0) {
-    unlink(cl_sock.sun_path);
+    int xerrno = errno;
+
+    (void) unlink(cl_sock.sun_path);
+    (void) close(sockfd);
     pr_signals_unblock();
+
+    errno = xerrno;
     return -1;
   }
 
 #if !defined(SO_PEERCRED) && !defined(HAVE_GETPEEREID) && \
     !defined(HAVE_GETPEERUCRED) && defined(LOCAL_CREDS)
   if (ctrls_connect_local_creds(sockfd) < 0) {
-    unlink(cl_sock.sun_path);
+    int xerrno = errno;
+
+    (void) unlink(cl_sock.sun_path);
+    (void) close(sockfd);
     pr_signals_unblock();
+
+    errno = xerrno;
     return -1;
   }
 #endif /* LOCAL_CREDS */
@@ -1027,10 +1221,15 @@ int pr_ctrls_issock_unix(mode_t sock_mode) {
 #if defined(SO_PEERCRED)
 static int ctrls_get_creds_peercred(int sockfd, uid_t *uid, gid_t *gid,
     pid_t *pid) {
+# if defined(HAVE_STRUCT_SOCKPEERCRED)
+  struct sockpeercred cred;
+# else
   struct ucred cred;
-  socklen_t credlen = sizeof(cred);
+# endif /* HAVE_STRUCT_SOCKPEERCRED */
+  socklen_t cred_len;
 
-  if (getsockopt(sockfd, SOL_SOCKET, SO_PEERCRED, &cred, &credlen) < 0) {
+  cred_len = sizeof(cred);
+  if (getsockopt(sockfd, SOL_SOCKET, SO_PEERCRED, &cred, &cred_len) < 0) {
     int xerrno = errno;
 
     pr_trace_msg(trace_channel, 2,
@@ -1041,14 +1240,17 @@ static int ctrls_get_creds_peercred(int sockfd, uid_t *uid, gid_t *gid,
     return -1;
   }
 
-  if (uid)
+  if (uid) {
     *uid = cred.uid;
+  }
 
-  if (gid)
+  if (gid) {
     *gid = cred.gid;
+  }
 
-  if (pid)
+  if (pid) {
     *pid = cred.pid;
+  }
 
   return 0;
 }
@@ -1529,7 +1731,7 @@ int pr_reset_ctrls(void) {
  */
 unsigned char pr_ctrls_check_group_acl(gid_t cl_gid,
     const ctrls_grp_acl_t *grp_acl) {
-  register int i = 0;
+  register unsigned int i = 0;
   unsigned char res = FALSE;
 
   /* Note: the special condition of ngids of 1 and gids of NULL signals
@@ -1542,11 +1744,13 @@ unsigned char pr_ctrls_check_group_acl(gid_t cl_gid,
       }
     }
 
-  } else if (grp_acl->ngids == 1)
+  } else if (grp_acl->ngids == 1) {
     res = TRUE;
+  }
 
-  if (!grp_acl->allow)
+  if (!grp_acl->allow) {
     res = !res;
+  }
 
   return res;
 }
@@ -1557,7 +1761,7 @@ unsigned char pr_ctrls_check_group_acl(gid_t cl_gid,
  */
 unsigned char pr_ctrls_check_user_acl(uid_t cl_uid,
     const ctrls_usr_acl_t *usr_acl) {
-  register int i = 0;
+  register unsigned int i = 0;
   unsigned char res = FALSE;
 
   /* Note: the special condition of nuids of 1 and uids of NULL signals
@@ -1570,11 +1774,13 @@ unsigned char pr_ctrls_check_user_acl(uid_t cl_uid,
       }
     }
 
-  } else if (usr_acl->nuids == 1)
+  } else if (usr_acl->nuids == 1) {
     res = TRUE;
+  }
 
-  if (!usr_acl->allow)
+  if (!usr_acl->allow) {
     res = !res;
+  }
 
   return res;
 }
@@ -1618,7 +1824,7 @@ static char *ctrls_argsep(char **arg) {
   if (!arg || !*arg || !**arg)
     return NULL;
 
-  while (**arg && isspace((int) **arg))
+  while (**arg && PR_ISSPACE(**arg))
     (*arg)++;
 
   if (!**arg)
@@ -1632,7 +1838,7 @@ static char *ctrls_argsep(char **arg) {
   }
 
   while (**arg && **arg != ',' &&
-      (quote_mode ? (**arg != '\"') : (!isspace((int) **arg)))) {
+      (quote_mode ? (**arg != '\"') : (!PR_ISSPACE(**arg)))) {
 
     if (**arg == '\\' && quote_mode) {
 
@@ -1955,19 +2161,18 @@ void init_ctrls(void) {
 
   sockfd = socket(AF_UNIX, SOCK_STREAM, 0);
   if (sockfd < 0) {
-    pr_log_pri(PR_LOG_DEBUG, "unable to create Unix domain socket: %s",
+    pr_log_debug(DEBUG10, "unable to create Unix domain socket: %s",
       strerror(errno));
     return;
   }
 
   memset(&sockun, 0, sizeof(sockun));
   sockun.sun_family = AF_UNIX;
-  sstrncpy(sockun.sun_path, sockpath, strlen(sockpath) + 1);
+  sstrncpy(sockun.sun_path, sockpath, sizeof(sockun.sun_path));
   socklen = sizeof(struct sockaddr_un);
 
   if (bind(sockfd, (struct sockaddr *) &sockun, socklen) < 0) {
-    pr_log_pri(PR_LOG_DEBUG,
-      "unable to bind to Unix domain socket at '%s': %s",
+    pr_log_debug(DEBUG10, "unable to bind to Unix domain socket at '%s': %s",
       sockpath, strerror(errno));
     (void) close(sockfd);
     (void) unlink(sockpath);
@@ -1975,8 +2180,7 @@ void init_ctrls(void) {
   }
 
   if (fstat(sockfd, &st) < 0) {
-    pr_log_pri(PR_LOG_DEBUG,
-      "unable to stat Unix domain socket at '%s': %s",
+    pr_log_debug(DEBUG10, "unable to stat Unix domain socket at '%s': %s",
       sockpath, strerror(errno));
     (void) close(sockfd);
     (void) unlink(sockpath);
