@@ -1,6 +1,6 @@
 /*
  * ProFTPD - FTP server daemon
- * Copyright (c) 2010-2013 The ProFTPD Project team
+ * Copyright (c) 2010-2016 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -22,14 +22,13 @@
  * OpenSSL in the source distribution.
  */
 
-/* Memcache management
- * $Id: memcache.c,v 1.26 2013-01-28 01:21:05 castaglia Exp $
- */
+/* Memcache management */
 
 #include "conf.h"
 
 #ifdef PR_USE_MEMCACHE
 
+#include "hanson-tpl.h"
 #include <libmemcached/memcached.h>
 
 #if defined(LIBMEMCACHED_VERSION_HEX)
@@ -182,14 +181,37 @@ static int mcache_set_options(pr_memcache_t *mcache, unsigned long flags,
 
   /* Use the binary protocol by default, unless explicitly requested not to. */
   val = memcached_behavior_get(mcache->mc, MEMCACHED_BEHAVIOR_BINARY_PROTOCOL);
+  pr_trace_msg(trace_channel, 16,
+    "found BINARY_PROTOCOL=%s default behavior (val %lu) for connection",
+    val != 1 ? "false" : "true", (unsigned long) val);
+
   if (val != 1) {
     if (!(flags & PR_MEMCACHE_FL_NO_BINARY_PROTOCOL)) {
       res = memcached_behavior_set(mcache->mc,
         MEMCACHED_BEHAVIOR_BINARY_PROTOCOL, 1);
       if (res != MEMCACHED_SUCCESS) {
         pr_trace_msg(trace_channel, 4,
-          "error setting BINARY_PROTOCOL behavior on connection: %s",
+          "error setting BINARY_PROTOCOL=true behavior on connection: %s",
           memcached_strerror(mcache->mc, res));
+
+      } else {
+        pr_trace_msg(trace_channel, 16, "%s",
+          "set BINARY_PROTOCOL=true for connection");
+      }
+    }
+
+  } else {
+    if (flags & PR_MEMCACHE_FL_NO_BINARY_PROTOCOL) {
+      res = memcached_behavior_set(mcache->mc,
+        MEMCACHED_BEHAVIOR_BINARY_PROTOCOL, 0);
+      if (res != MEMCACHED_SUCCESS) {
+        pr_trace_msg(trace_channel, 4,
+          "error setting BINARY_PROTOCOL=false behavior on connection: %s",
+          memcached_strerror(mcache->mc, res));
+
+      } else {
+        pr_trace_msg(trace_channel, 16, "%s",
+          "set BINARY_PROTOCOL=false for connection");
       }
     }
   }
@@ -285,10 +307,18 @@ static int mcache_ping_servers(pr_memcache_t *mcache) {
   }
 
   memcached_servers_reset(clone);
+
+  /* Bug#4242: Don't use memcached_server_push() if we're using
+   * libmemcached-1.0.18 or earlier.  Doing so leads to a segfault, due to
+   * this libmemcached bug:
+   *
+   *  https://bugs.launchpad.net/libmemcached/+bug/1154159
+   */
+#if LIBMEMCACHED_VERSION_HEX > 0x01000018
   memcached_server_push(clone, configured_server_list);
+#endif
 
   server_count = memcached_server_count(clone);
-
   pr_trace_msg(trace_channel, 16,
     "pinging %lu memcached %s", (unsigned long) server_count,
     server_count != 1 ? "servers" : "server");
@@ -404,7 +434,6 @@ static int mcache_stat_servers(pr_memcache_t *mcache) {
             pr_trace_msg(trace_channel, 3,
               "error requesting memcached stats: system error: %s",
               strerror(errno));
-            break;
 
           } else {
             /* We know that we're not using nonblocking IO; this value usually
@@ -414,6 +443,7 @@ static int mcache_stat_servers(pr_memcache_t *mcache) {
              */
             res = MEMCACHED_CONNECTION_FAILURE;
           }
+          break;
 
           case MEMCACHED_SOME_ERRORS:
           case MEMCACHED_SERVER_MARKED_DEAD:
@@ -426,6 +456,8 @@ static int mcache_stat_servers(pr_memcache_t *mcache) {
                 "unable to connect to %s:%d", memcached_server_name(server),
                 memcached_server_port(server));
             }
+
+            break;
           }
 
         default:
@@ -491,10 +523,10 @@ pr_memcache_t *pr_memcache_conn_new(pool *p, module *m, unsigned long flags,
     return NULL;
   }
 
-  sub_pool = pr_pool_create_sz(p, 128);
+  sub_pool = make_sub_pool(p);
   pr_pool_tag(sub_pool, "Memcache connection pool");
 
-  mcache = palloc(sub_pool, sizeof(pr_memcache_t));
+  mcache = pcalloc(sub_pool, sizeof(pr_memcache_t));
   mcache->pool = sub_pool;
   mcache->owner = m;
   mcache->mc = mc;
@@ -568,6 +600,31 @@ int pr_memcache_conn_close(pr_memcache_t *mcache) {
   return 0;
 }
 
+int pr_memcache_conn_clone(pool *p, pr_memcache_t *mcache) {
+  memcached_st *old_mc = NULL, *new_mc = NULL;
+
+  if (p == NULL ||
+      mcache == NULL) {
+    errno = EINVAL;
+    return -1;
+  }
+
+  memcached_quit(mcache->mc);
+  old_mc = mcache->mc;
+
+  new_mc = memcached_clone(NULL, old_mc);
+  if (new_mc == NULL) {
+    errno = ENOMEM;
+    return -1;
+  }
+
+  /* Now free up the previous context; we don't need it anymore. */
+  memcached_free(old_mc);
+
+  mcache->mc = new_mc;
+  return 0;
+}
+
 static int modptr_cmp_cb(const void *k1, size_t ksz1, const void *k2,
     size_t ksz2) {
 
@@ -856,16 +913,15 @@ static void mcache_set_module_namespace(pr_memcache_t *mcache, module *m) {
 
   } else {
     if (mcache->namespace_tab != NULL) {
-      void *v;
+      const char *v;
 
       v = pr_table_kget(mcache->namespace_tab, m, sizeof(module *), NULL);
-      if (v) {
+      if (v != NULL) {
         pr_trace_msg(trace_channel, 25,
-          "using namespace prefix '%s' for module 'mod_%s.c'", (const char *) v,
-          m->name);
+          "using namespace prefix '%s' for module 'mod_%s.c'", v, m->name);
 
         res = memcached_callback_set(mcache->mc, MEMCACHED_CALLBACK_PREFIX_KEY,
-          v);
+          (void *) v);
       }
 
     } else {
@@ -874,9 +930,16 @@ static void mcache_set_module_namespace(pr_memcache_t *mcache, module *m) {
   }
 
   if (res != MEMCACHED_SUCCESS) {
-    pr_trace_msg(trace_channel, 9,
-      "unable to set MEMCACHED_CALLBACK_PREFIX_KEY for module 'mod_%s.c': %s",
-      m->name, memcached_strerror(mcache->mc, res));
+    if (m != NULL) {
+      pr_trace_msg(trace_channel, 9,
+        "unable to set MEMCACHED_CALLBACK_PREFIX_KEY for module 'mod_%s.c': %s",
+        m->name, memcached_strerror(mcache->mc, res));
+
+    } else {
+      pr_trace_msg(trace_channel, 9,
+        "unable to clear MEMCACHED_CALLBACK_PREFIX_KEY: %s",
+        memcached_strerror(mcache->mc, res));
+    }
   }
 }
 
@@ -912,7 +975,6 @@ int pr_memcache_kadd(pr_memcache_t *mcache, module *m, const char *key,
           (unsigned long) keysz, (unsigned long) valuesz, strerror(xerrno));
 
         errno = xerrno;
-        break;
 
       } else {
         /* We know that we're not using nonblocking IO; this value usually
@@ -922,6 +984,7 @@ int pr_memcache_kadd(pr_memcache_t *mcache, module *m, const char *key,
          */
         res = MEMCACHED_CONNECTION_FAILURE;
       }
+      break;
 
     case MEMCACHED_SERVER_MARKED_DEAD:
     case MEMCACHED_CONNECTION_FAILURE: {
@@ -933,6 +996,8 @@ int pr_memcache_kadd(pr_memcache_t *mcache, module *m, const char *key,
           "unable to connect to %s:%d", memcached_server_name(server),
           memcached_server_port(server));
       }
+
+      break;
     }
 
     default:
@@ -941,6 +1006,7 @@ int pr_memcache_kadd(pr_memcache_t *mcache, module *m, const char *key,
         (unsigned long) keysz, (unsigned long) valuesz,
         memcached_strerror(mcache->mc, res));
       errno = EPERM;
+      break;
   }
 
   return -1;
@@ -958,6 +1024,10 @@ int pr_memcache_kdecr(pr_memcache_t *mcache, module *m, const char *key,
     return -1;
   }
 
+  /* Note: libmemcached automatically handles the case where value might be
+   * NULL.
+   */
+
   mcache_set_module_namespace(mcache, m);
   res = memcached_decrement(mcache->mc, key, keysz, decr, value);
   mcache_set_module_namespace(mcache, NULL);
@@ -975,7 +1045,6 @@ int pr_memcache_kdecr(pr_memcache_t *mcache, module *m, const char *key,
           (unsigned long) keysz, (unsigned long) decr, strerror(xerrno));
 
         errno = xerrno;
-        break;
 
       } else {
         /* We know that we're not using nonblocking IO; this value usually
@@ -985,6 +1054,7 @@ int pr_memcache_kdecr(pr_memcache_t *mcache, module *m, const char *key,
          */
         res = MEMCACHED_CONNECTION_FAILURE;
       }
+      break;
 
     case MEMCACHED_SERVER_MARKED_DEAD:
     case MEMCACHED_CONNECTION_FAILURE: {
@@ -996,6 +1066,8 @@ int pr_memcache_kdecr(pr_memcache_t *mcache, module *m, const char *key,
           "unable to connect to %s:%d", memcached_server_name(server),
           memcached_server_port(server));
       }
+
+      break;
     }
 
     default:
@@ -1004,6 +1076,7 @@ int pr_memcache_kdecr(pr_memcache_t *mcache, module *m, const char *key,
         (unsigned long) keysz, (unsigned long) decr,
         memcached_strerror(mcache->mc, res));
       errno = EPERM;
+      break;
   }
 
   return -1;
@@ -1014,6 +1087,7 @@ void *pr_memcache_kget(pr_memcache_t *mcache, module *m, const char *key,
   char *data = NULL;
   void *ptr = NULL;
   memcached_return res;
+  int xerrno = 0;
 
   if (mcache == NULL ||
       m == NULL ||
@@ -1026,6 +1100,7 @@ void *pr_memcache_kget(pr_memcache_t *mcache, module *m, const char *key,
 
   mcache_set_module_namespace(mcache, m);
   data = memcached_get(mcache->mc, key, keysz, valuesz, flags, &res);
+  xerrno = errno;
   mcache_set_module_namespace(mcache, NULL);
 
   if (data == NULL) {
@@ -1038,13 +1113,11 @@ void *pr_memcache_kget(pr_memcache_t *mcache, module *m, const char *key,
 
       case MEMCACHED_ERRNO:
         if (errno != EINPROGRESS) {
-          int xerrno = errno;
           pr_trace_msg(trace_channel, 3,
             "no data found for key (%lu bytes): system error: %s",
             (unsigned long) keysz, strerror(xerrno));
 
           errno = xerrno;
-          break;
 
         } else {
           /* We know that we're not using nonblocking IO; this value usually
@@ -1054,6 +1127,7 @@ void *pr_memcache_kget(pr_memcache_t *mcache, module *m, const char *key,
            */
           res = MEMCACHED_CONNECTION_FAILURE;
         }
+        break;
 
       case MEMCACHED_SERVER_MARKED_DEAD:
       case MEMCACHED_CONNECTION_FAILURE: {
@@ -1065,7 +1139,9 @@ void *pr_memcache_kget(pr_memcache_t *mcache, module *m, const char *key,
             "unable to connect to %s:%d", memcached_server_name(server),
             memcached_server_port(server));
         }
-      } 
+
+        break;
+      }
 
       default:
         pr_trace_msg(trace_channel, 6,
@@ -1094,6 +1170,7 @@ char *pr_memcache_kget_str(pr_memcache_t *mcache, module *m, const char *key,
   char *data = NULL, *ptr = NULL;
   size_t valuesz = 0;
   memcached_return res;
+  int xerrno = 0;
 
   if (mcache == NULL ||
       m == NULL ||
@@ -1105,6 +1182,7 @@ char *pr_memcache_kget_str(pr_memcache_t *mcache, module *m, const char *key,
 
   mcache_set_module_namespace(mcache, m);
   data = memcached_get(mcache->mc, key, keysz, &valuesz, flags, &res);
+  xerrno = errno;
   mcache_set_module_namespace(mcache, NULL);
 
   if (data == NULL) {
@@ -1117,14 +1195,11 @@ char *pr_memcache_kget_str(pr_memcache_t *mcache, module *m, const char *key,
 
       case MEMCACHED_ERRNO:
         if (errno != EINPROGRESS) {
-          int xerrno = errno;
-
           pr_trace_msg(trace_channel, 3,
             "no data found for key (%lu bytes): system error: %s",
             (unsigned long) keysz, strerror(xerrno));
 
           errno = xerrno;
-          break;
 
         } else {
           /* We know that we're not using nonblocking IO; this value usually
@@ -1134,6 +1209,7 @@ char *pr_memcache_kget_str(pr_memcache_t *mcache, module *m, const char *key,
            */
           res = MEMCACHED_CONNECTION_FAILURE;
         }
+        break;
 
       case MEMCACHED_SERVER_MARKED_DEAD:
       case MEMCACHED_CONNECTION_FAILURE: {
@@ -1145,6 +1221,8 @@ char *pr_memcache_kget_str(pr_memcache_t *mcache, module *m, const char *key,
             "unable to connect to %s:%d", memcached_server_name(server),
             memcached_server_port(server));
         }
+
+        break;
       }
 
       default:
@@ -1181,6 +1259,10 @@ int pr_memcache_kincr(pr_memcache_t *mcache, module *m, const char *key,
     return -1;
   }
 
+  /* Note: libmemcached automatically handles the case where value might be
+   * NULL.
+   */
+
   mcache_set_module_namespace(mcache, m);
   res = memcached_increment(mcache->mc, key, keysz, incr, value);
   mcache_set_module_namespace(mcache, NULL);
@@ -1208,7 +1290,6 @@ int pr_memcache_kincr(pr_memcache_t *mcache, module *m, const char *key,
           (unsigned long) keysz, (unsigned long) incr, strerror(xerrno));
 
         errno = xerrno;
-        break;
 
       } else {
         /* We know that we're not using nonblocking IO; this value usually
@@ -1218,6 +1299,7 @@ int pr_memcache_kincr(pr_memcache_t *mcache, module *m, const char *key,
          */
         res = MEMCACHED_CONNECTION_FAILURE;
       }
+      break;
 
     case MEMCACHED_SERVER_MARKED_DEAD:
     case MEMCACHED_CONNECTION_FAILURE: {
@@ -1229,6 +1311,8 @@ int pr_memcache_kincr(pr_memcache_t *mcache, module *m, const char *key,
           "unable to connect to %s:%d", memcached_server_name(server),
           memcached_server_port(server));
       }
+
+      break;
     }
 
     default:
@@ -1237,6 +1321,7 @@ int pr_memcache_kincr(pr_memcache_t *mcache, module *m, const char *key,
         (unsigned long) keysz, (unsigned long) incr,
         memcached_strerror(mcache->mc, res));
       errno = EPERM;
+      break;
   }
 
   return -1;
@@ -1270,7 +1355,6 @@ int pr_memcache_kremove(pr_memcache_t *mcache, module *m, const char *key,
           (unsigned long) keysz, strerror(xerrno));
 
         errno = xerrno;
-        break;
 
       } else {
         /* We know that we're not using nonblocking IO; this value usually
@@ -1280,6 +1364,7 @@ int pr_memcache_kremove(pr_memcache_t *mcache, module *m, const char *key,
          */
         res = MEMCACHED_CONNECTION_FAILURE;
       }
+      break;
 
     case MEMCACHED_SERVER_MARKED_DEAD:
     case MEMCACHED_CONNECTION_FAILURE: {
@@ -1291,6 +1376,8 @@ int pr_memcache_kremove(pr_memcache_t *mcache, module *m, const char *key,
           "unable to connect to %s:%d", memcached_server_name(server),
           memcached_server_port(server));
       }
+
+      break;
     }
 
     default:
@@ -1298,6 +1385,7 @@ int pr_memcache_kremove(pr_memcache_t *mcache, module *m, const char *key,
         "error removing key (%lu bytes): %s", (unsigned long) keysz,
         memcached_strerror(mcache->mc, res));
       errno = EPERM;
+      break;
   }
 
   return -1;
@@ -1319,7 +1407,8 @@ int pr_memcache_kset(pr_memcache_t *mcache, module *m, const char *key,
   }
 
   mcache_set_module_namespace(mcache, m);
-  res = memcached_set(mcache->mc, key, keysz, value, valuesz, expires, flags); 
+  res = memcached_set(mcache->mc, key, keysz, (const char *) value, valuesz,
+    expires, flags);
   mcache_set_module_namespace(mcache, NULL);
 
   switch (res) {
@@ -1335,7 +1424,6 @@ int pr_memcache_kset(pr_memcache_t *mcache, module *m, const char *key,
           (unsigned long) keysz, (unsigned long) valuesz, strerror(xerrno));
 
         errno = xerrno;
-        break;
 
       } else {
         /* We know that we're not using nonblocking IO; this value usually
@@ -1345,6 +1433,7 @@ int pr_memcache_kset(pr_memcache_t *mcache, module *m, const char *key,
          */
         res = MEMCACHED_CONNECTION_FAILURE;
       }
+      break;
 
     case MEMCACHED_SERVER_MARKED_DEAD:
     case MEMCACHED_CONNECTION_FAILURE: {
@@ -1356,6 +1445,8 @@ int pr_memcache_kset(pr_memcache_t *mcache, module *m, const char *key,
           "unable to connect to %s:%d", memcached_server_name(server),
           memcached_server_port(server));
       }
+
+      break;
     }
 
     default:
@@ -1364,6 +1455,7 @@ int pr_memcache_kset(pr_memcache_t *mcache, module *m, const char *key,
         (unsigned long) keysz, (unsigned long) valuesz,
         memcached_strerror(mcache->mc, res));
       errno = EPERM;
+      break;
   }
 
   return -1;
@@ -1470,6 +1562,11 @@ int pr_memcache_conn_close(pr_memcache_t *mcache) {
   return -1;
 }
 
+int pr_memcache_conn_clone(pool *p, pr_memcache_t *mcache) {
+  errno = ENOSYS;
+  return -1;
+}
+
 int pr_memcache_conn_set_namespace(pr_memcache_t *mcache, module *m,
     const char *prefix) {
   errno = ENOSYS;
