@@ -1,6 +1,6 @@
 /*
  * ProFTPD - FTP server daemon
- * Copyright (c) 2007-2012 The ProFTPD Project team
+ * Copyright (c) 2007-2016 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -22,9 +22,7 @@
  * OpenSSL in the source distribution.
  */
 
-/* Proctitle management
- * $Id: proctitle.c,v 1.13 2012-04-15 18:04:15 castaglia Exp $
- */
+/* Proctitle management */
 
 #include "conf.h"
 
@@ -52,10 +50,8 @@ extern char *__progname, *__progname_full;
 #endif /* HAVE___PROGNAME */
 extern char **environ;
 
-#ifndef PR_DEVEL_STACK_TRACE
 static char **prog_argv = NULL;
 static char *prog_last_argv = NULL;
-#endif /* PR_DEVEL_STACK_TRACE */
 
 static int prog_argc = -1;
 static char proc_title_buf[BUFSIZ];
@@ -64,16 +60,17 @@ static unsigned int proc_flags = 0;
 #define PR_PROCTITLE_FL_USE_STATIC		0x001
 
 void pr_proctitle_init(int argc, char *argv[], char *envp[]) {
-#ifndef PR_DEVEL_STACK_TRACE
   register int i;
   register size_t envpsize;
   char **p;
 
   /* Move the environment so setproctitle can use the space. */
-  for (i = envpsize = 0; envp[i] != NULL; i++)
+  for (i = envpsize = 0; envp[i] != NULL; i++) {
     envpsize += strlen(envp[i]) + 1;
+  }
 
-  if ((p = (char **) malloc((i + 1) * sizeof(char *))) != NULL) {
+  p = (char **) malloc((i + 1) * sizeof(char *));
+  if (p != NULL) {
     environ = p;
 
     for (i = 0; envp[i] != NULL; i++) {
@@ -97,63 +94,56 @@ void pr_proctitle_init(int argc, char *argv[], char *envp[]) {
   prog_argc = argc;
 
   for (i = 0; i < prog_argc; i++) {
-    if (!i || (prog_last_argv + 1 == argv[i]))
+    if (!i || (prog_last_argv + 1 == argv[i])) {
       prog_last_argv = argv[i] + strlen(argv[i]);
+    }
   }
 
   for (i = 0; envp[i] != NULL; i++) {
-    if ((prog_last_argv + 1) == envp[i])
+    if ((prog_last_argv + 1) == envp[i]) {
       prog_last_argv = envp[i] + strlen(envp[i]);
+    }
   }
 
-# ifdef HAVE___PROGNAME
+#ifdef HAVE___PROGNAME
   /* Set the __progname and __progname_full variables so glibc and company
    * don't go nuts.
    */
   __progname = strdup("proftpd");
   __progname_full = strdup(argv[0]);
-# endif /* HAVE___PROGNAME */
-#else
-  /* Silence compiler warning about unused variable when stacktrace
-   * developer mode is configured.
-   */
-  prog_argc = -1;
-#endif /* !PR_DEVEL_STACK_TRACE */
+#endif /* HAVE___PROGNAME */
   memset(proc_title_buf, '\0', sizeof(proc_title_buf));
 }
 
 void pr_proctitle_free(void) {
 #ifdef PR_USE_DEVEL
-# ifndef PR_DEVEL_STACK_TRACE
   if (environ) {
     register unsigned int i;
 
-    for (i = 0; environ[i] != NULL; i++)
+    for (i = 0; environ[i] != NULL; i++) {
       free(environ[i]);
+    }
     free(environ);
     environ = NULL;
   }
 
-#  ifdef HAVE___PROGNAME
+# ifdef HAVE___PROGNAME
   free(__progname);
   __progname = NULL;
   free(__progname_full);
   __progname_full = NULL;
-#  endif /* HAVE___PROGNAME */
-# endif /* !PR_DEVEL_STACK_TRACE */
+# endif /* HAVE___PROGNAME */
 #endif /* PR_USE_DEVEL */
 }
 
 void pr_proctitle_set_str(const char *str) {
-#ifndef PR_DEVEL_STACK_TRACE
-
-# ifndef HAVE_SETPROCTITLE
+#ifndef HAVE_SETPROCTITLE
   char *p;
   int i, procbuflen, maxlen = (prog_last_argv - prog_argv[0]) - 2;
 
-#  if PF_ARGV_TYPE == PF_ARGV_PSTAT
+# if PF_ARGV_TYPE == PF_ARGV_PSTAT
   union pstun pst;
-#  endif /* PF_ARGV_PSTAT */
+# endif /* PF_ARGV_PSTAT */
 
   if (proc_flags & PR_PROCTITLE_FL_USE_STATIC) {
     return;
@@ -162,130 +152,129 @@ void pr_proctitle_set_str(const char *str) {
   sstrncpy(proc_title_buf, str, sizeof(proc_title_buf));
   procbuflen = strlen(proc_title_buf);
 
-#  if PF_ARGV_TYPE == PF_ARGV_NEW
+# if PF_ARGV_TYPE == PF_ARGV_NEW
   /* We can just replace argv[] arguments.  Nice and easy. */
   prog_argv[0] = proc_title_buf;
   for (i = 1; i < prog_argc; i++) {
     prog_argv[i] = "";
   }
-#  endif /* PF_ARGV_NEW */
+# endif /* PF_ARGV_NEW */
 
-#  if PF_ARGV_TYPE == PF_ARGV_WRITEABLE
+# if PF_ARGV_TYPE == PF_ARGV_WRITEABLE
   /* We can overwrite individual argv[] arguments.  Semi-nice. */
   snprintf(prog_argv[0], maxlen, "%s", proc_title_buf);
   p = &prog_argv[0][procbuflen];
 
-  while (p < prog_last_argv)
+  while (p < prog_last_argv) {
     *p++ = '\0';
+  }
 
   for (i = 1; i < prog_argc; i++) {
     prog_argv[i] = "";
   }
 
-#  endif /* PF_ARGV_WRITEABLE */
+# endif /* PF_ARGV_WRITEABLE */
 
-#  if PF_ARGV_TYPE == PF_ARGV_PSSTRINGS
+# if PF_ARGV_TYPE == PF_ARGV_PSSTRINGS
   PS_STRINGS->ps_nargvstr = 1;
   PS_STRINGS->ps_argvstr = proc_title_buf;
-#  endif /* PF_ARGV_PSSTRINGS */
+# endif /* PF_ARGV_PSSTRINGS */
 
-# else
+#else
   if (proc_flags & PR_PROCTITLE_FL_USE_STATIC) {
     return;
   }
 
   setproctitle("%s", str);
-# endif /* HAVE_SETPROCTITLE */
-#endif /* PR_DEVEL_STACK_TRACE */
+#endif /* HAVE_SETPROCTITLE */
 }
 
 void pr_proctitle_set(const char *fmt, ...) {
-#ifndef PR_DEVEL_STACK_TRACE
   va_list msg;
 
-# ifndef HAVE_SETPROCTITLE
-#  if PF_ARGV_TYPE == PF_ARGV_PSTAT
+#ifndef HAVE_SETPROCTITLE
+# if PF_ARGV_TYPE == PF_ARGV_PSTAT
   union pstun pst;
-#  endif /* PF_ARGV_PSTAT */
+# endif /* PF_ARGV_PSTAT */
   char *p;
   int i, procbuflen, maxlen = (prog_last_argv - prog_argv[0]) - 2;
-# endif /* HAVE_SETPROCTITLE */
+#endif /* HAVE_SETPROCTITLE */
 
   if (proc_flags & PR_PROCTITLE_FL_USE_STATIC) {
     return;
   }
 
-  if (!fmt)
+  if (fmt == NULL) {
     return;
+  }
 
   va_start(msg, fmt);
 
   memset(proc_title_buf, 0, sizeof(proc_title_buf));
 
-# ifdef HAVE_SETPROCTITLE
-#  if __FreeBSD__ >= 4 && !defined(FREEBSD4_0) && !defined(FREEBSD4_1)
+#ifdef HAVE_SETPROCTITLE
+# if __FreeBSD__ >= 4 && !defined(FREEBSD4_0) && !defined(FREEBSD4_1)
   /* FreeBSD's setproctitle() automatically prepends the process name. */
   vsnprintf(proc_title_buf, sizeof(proc_title_buf), fmt, msg);
 
-#  else /* FREEBSD4 */
+# else /* FREEBSD4 */
   /* Manually append the process name for non-FreeBSD platforms. */
   snprintf(proc_title_buf, sizeof(proc_title_buf), "%s", "proftpd: ");
   vsnprintf(proc_title_buf + strlen(proc_title_buf),
     sizeof(proc_title_buf) - strlen(proc_title_buf), fmt, msg);
 
-#  endif /* FREEBSD4 */
+# endif /* FREEBSD4 */
   setproctitle("%s", proc_title_buf);
 
-# else /* HAVE_SETPROCTITLE */
+#else /* HAVE_SETPROCTITLE */
   /* Manually append the process name for non-setproctitle() platforms. */
   snprintf(proc_title_buf, sizeof(proc_title_buf), "%s", "proftpd: ");
   vsnprintf(proc_title_buf + strlen(proc_title_buf),
     sizeof(proc_title_buf) - strlen(proc_title_buf), fmt, msg);
 
-# endif /* HAVE_SETPROCTITLE */
+#endif /* HAVE_SETPROCTITLE */
 
   va_end(msg);
 
-# ifdef HAVE_SETPROCTITLE
+#ifdef HAVE_SETPROCTITLE
   return;
-# else
+#else
   procbuflen = strlen(proc_title_buf);
 
-#  if PF_ARGV_TYPE == PF_ARGV_NEW
+# if PF_ARGV_TYPE == PF_ARGV_NEW
   /* We can just replace argv[] arguments.  Nice and easy. */
   prog_argv[0] = proc_title_buf;
   for (i = 1; i < prog_argc; i++) {
     prog_argv[i] = "";
   }
-#  endif /* PF_ARGV_NEW */
+# endif /* PF_ARGV_NEW */
 
-#  if PF_ARGV_TYPE == PF_ARGV_WRITEABLE
+# if PF_ARGV_TYPE == PF_ARGV_WRITEABLE
   /* We can overwrite individual argv[] arguments.  Semi-nice. */
   snprintf(prog_argv[0], maxlen, "%s", proc_title_buf);
   p = &prog_argv[0][procbuflen];
 
-  while (p < prog_last_argv)
+  while (p < prog_last_argv) {
     *p++ = '\0';
+  }
 
   for (i = 1; i < prog_argc; i++) {
     prog_argv[i] = "";
   }
 
-#  endif /* PF_ARGV_WRITEABLE */
+# endif /* PF_ARGV_WRITEABLE */
 
-#  if PF_ARGV_TYPE == PF_ARGV_PSTAT
+# if PF_ARGV_TYPE == PF_ARGV_PSTAT
   pst.pst_command = proc_title_buf;
   pstat(PSTAT_SETCMD, pst, procbuflen, 0, 0);
 
-#  endif /* PF_ARGV_PSTAT */
+# endif /* PF_ARGV_PSTAT */
 
-#  if PF_ARGV_TYPE == PF_ARGV_PSSTRINGS
+# if PF_ARGV_TYPE == PF_ARGV_PSSTRINGS
   PS_STRINGS->ps_nargvstr = 1;
   PS_STRINGS->ps_argvstr = proc_title_buf;
-#  endif /* PF_ARGV_PSSTRINGS */
-
-# endif /* HAVE_SETPROCTITLE */
-#endif /* !PR_DEVEL_STACK_TRACE */
+# endif /* PF_ARGV_PSSTRINGS */
+#endif /* HAVE_SETPROCTITLE */
 }
 
 void pr_proctitle_set_static_str(const char *buf) {
