@@ -1,6 +1,6 @@
 /*
  * ProFTPD - FTP server daemon
- * Copyright (c) 2004 The ProFTPD Project team
+ * Copyright (c) 2004-2016 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -14,7 +14,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
  *
  * As a special exemption, The ProFTPD Project team and other respective
  * copyright holders give permission to link this program with OpenSSL, and
@@ -22,9 +22,7 @@
  * OpenSSL in the source distribution.
  */
 
-/* Variables API implementation
- * $Id: var.c,v 1.4 2006/12/11 23:11:34 castaglia Exp $
- */
+/* Variables API implementation */
 
 #include "conf.h"
 
@@ -41,17 +39,19 @@ static pr_table_t *var_tab = NULL;
 
 typedef const char *(*var_vstr_cb)(void *, size_t);
 
+static const char *trace_channel = "var";
+
 /* Public API
  */
 
 int pr_var_delete(const char *name) {
-  if (!name) {
-    errno = EINVAL;
+  if (var_tab == NULL) {
+    errno = EPERM;
     return -1;
   }
 
-  if (!var_tab) {
-    errno = EPERM;
+  if (name == NULL) {
+    errno = EINVAL;
     return -1;
   }
 
@@ -59,13 +59,13 @@ int pr_var_delete(const char *name) {
 }
 
 int pr_var_exists(const char *name) {
-  if (!name) {
-    errno = EINVAL;
+  if (var_tab == NULL) {
+    errno = EPERM;
     return -1;
   }
 
-  if (!var_tab) {
-    errno = EPERM;
+  if (name == NULL) {
+    errno = EINVAL;
     return -1;
   }
 
@@ -73,21 +73,22 @@ int pr_var_exists(const char *name) {
 }
 
 const char *pr_var_get(const char *name) {
-  struct var *v;
+  const struct var *v = NULL;
 
-  if (!name) {
-    errno = EINVAL;
-    return NULL;
-  }
-
-  if (!var_tab) {
+  if (var_tab == NULL) {
     errno = EPERM;
     return NULL;
   }
 
-  v = pr_table_get(var_tab, name, NULL);
-  if (!v)
+  if (name == NULL) {
+    errno = EINVAL;
     return NULL;
+  }
+
+  v = pr_table_get(var_tab, name, NULL);
+  if (v == NULL) {
+    return NULL;
+  }
 
   switch (v->v_type) {
     case PR_VAR_TYPE_STR:
@@ -100,7 +101,8 @@ const char *pr_var_get(const char *name) {
 
     default:
       /* Pass through to the error case. */
-      ;
+      pr_trace_msg(trace_channel, 9,
+        "unknown var type (%d) found for name '%s'", v->v_type, name);
   }
 
   errno = EINVAL;
@@ -109,61 +111,83 @@ const char *pr_var_get(const char *name) {
 
 const char *pr_var_next(const char **desc) {
   const char *name;
-  struct var *v;
+  const struct var *v;
 
-  if (!var_tab) {
+  if (var_tab == NULL) {
     errno = EPERM;
     return NULL;
   }
 
   name = pr_table_next(var_tab);
-  if (!name)
+  if (name == NULL) {
     return NULL;
+  }
 
   v = pr_table_get(var_tab, name, NULL);
-  if (v && desc)
+  if (v != NULL &&
+      desc != NULL) {
     *desc = v->v_desc;
+  }
 
   return name;
 }
 
-int pr_var_remove(const char *name) {
-  if (!name) {
-    errno = EINVAL;
-    return -1;
+void pr_var_rewind(void) {
+  if (var_tab != NULL) {
+    pr_table_rewind(var_tab);
   }
+}
 
-  if (!var_tab) {
+int pr_var_set(pool *p, const char *name, const char *desc, int vtype,
+    void *val, void *data, size_t datasz) {
+  struct var *v;
+  size_t namelen = 0;
+
+  if (var_tab == NULL) {
     errno = EPERM;
     return -1;
   }
 
-  return pr_table_remove(var_tab, name, NULL) ? 0 : -1;
-}
+  if (p == NULL ||
+      name == NULL ||
+      val == NULL) {
+    errno = EINVAL;
+    return -1;
+  }
 
-void pr_var_rewind(void) {
-  if (var_tab)
-    pr_table_rewind(var_tab);
-}
+  /* The length of the key must be greater than 3 characters (for "%{}"). */
+  namelen = strlen(name);
+  if (namelen < 4) {
+    errno = EINVAL;
+    return -1;
+  }
 
-int pr_var_set(pool *p, const char *name, const char *desc, int type,
-    void *val, void *data, size_t datasz) {
-  struct var *v;
-
-  if (!p || !name || !val) {
+  /* If the given variable type is not recognized, reject. */
+  if (vtype != PR_VAR_TYPE_STR &&
+      vtype != PR_VAR_TYPE_FUNC) {
     errno = EINVAL;
     return -1;
   }
 
   /* Specifying data, but no length for that data, is an error. */
-  if (data && datasz == 0) {
+  if (data != NULL &&
+      datasz == 0) {
+    errno = EINVAL;
+    return -1;
+  }
+
+  /* Specifying no data, but providing a non-zero length for that data, is an
+   * error.
+   */
+  if (data == NULL &&
+      datasz > 0) {
     errno = EINVAL;
     return -1;
   }
 
   /* Variable names MUST start with '%{', and end in '}'. */
   if (strncmp(name, "%{", 2) != 0 ||
-      name[strlen(name)-1] != '}') {
+      name[namelen-1] != '}') {
     errno = EINVAL;
     return -1;
   }
@@ -184,12 +208,14 @@ int pr_var_set(pool *p, const char *name, const char *desc, int type,
    * growth.
    */
 
-  switch (type) {
+  switch (vtype) {
     case PR_VAR_TYPE_STR:
       v = pcalloc(p, sizeof(struct var));
 
-      if (desc)
+      if (desc) {
         v->v_desc = (const char *) pstrdup(p, desc);
+      }
+
       v->v_type = PR_VAR_TYPE_STR; 
       v->v_val = pstrdup(p, (char *) val);
       v->v_datasz = strlen((char *) val);
@@ -198,10 +224,12 @@ int pr_var_set(pool *p, const char *name, const char *desc, int type,
     case PR_VAR_TYPE_FUNC:
       v = pcalloc(p, sizeof(struct var));
 
-      if (desc)
+      if (desc) {
         v->v_desc = (const char *) pstrdup(p, desc);
+      }
+
       v->v_type = PR_VAR_TYPE_FUNC; 
-      v->v_val = (var_vstr_cb) val;
+      v->v_val = val;
 
       if (data) {
         v->v_data = data;
@@ -220,13 +248,14 @@ int pr_var_set(pool *p, const char *name, const char *desc, int type,
 
 int var_init(void) {
 
-  if (!var_pool) {
+  if (var_pool == NULL) {
     var_pool = make_sub_pool(permanent_pool);
     pr_pool_tag(var_pool, "Variables Pool");
   }
 
-  if (!var_tab)
+  if (var_tab == NULL) {
     var_tab = pr_table_alloc(var_pool, 0);
+  }
 
   return 0;
 }
