@@ -1,6 +1,6 @@
 /*
  * ProFTPD - FTP server API testsuite
- * Copyright (c) 2011-2014 The ProFTPD Project team
+ * Copyright (c) 2011-2016 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -22,9 +22,7 @@
  * OpenSSL in the source distribution.
  */
 
-/* Command API tests
- * $Id: cmd.c,v 1.4 2014-01-27 18:25:15 castaglia Exp $
- */
+/* Command API tests */
 
 #include "tests.h"
 
@@ -264,6 +262,15 @@ START_TEST (cmd_get_id_test) {
   fail_unless(res == PR_CMD_LANG_ID, "Expected cmd ID %d for %s, got %d",
     PR_CMD_LANG_ID, C_LANG, res); 
 
+  res = pr_cmd_get_id(C_HOST);
+  fail_unless(res == PR_CMD_HOST_ID, "Expected cmd ID %d for %s, got %d",
+    PR_CMD_HOST_ID, C_HOST, res); 
+
+  res = pr_cmd_get_id(C_CLNT);
+  fail_unless(res == PR_CMD_CLNT_ID, "Expected cmd ID %d for %s, got %d",
+    PR_CMD_CLNT_ID, C_CLNT, res); 
+
+  /* RFC 2228 commands */
   res = pr_cmd_get_id(C_ADAT);
   fail_unless(res == PR_CMD_ADAT_ID, "Expected cmd ID %d for %s, got %d",
     PR_CMD_ADAT_ID, C_ADAT, res); 
@@ -357,8 +364,9 @@ START_TEST (cmd_strcmp_test) {
 END_TEST
 
 START_TEST (cmd_get_displayable_str_test) {
-  char *ok, *res = NULL;
+  const char *ok, *res = NULL;
   cmd_rec *cmd = NULL;
+  size_t len = 0;
 
   res = pr_cmd_get_displayable_str(NULL, NULL);
   fail_unless(res == NULL, "Failed to handle null cmd_rec");
@@ -383,6 +391,11 @@ START_TEST (cmd_get_displayable_str_test) {
    */
   fail_unless(strcmp(res, ok) == 0, "Expected '%s', got '%s'", ok, res);
 
+  if (pr_cmd_clear_cache(NULL) < 0) {
+    fail_unless(errno == EINVAL, "Expected EINVAL (%d), got %s (%d)", EINVAL,
+      strerror(errno), errno);
+  }
+
   mark_point();
   pr_cmd_clear_cache(cmd);
   res = pr_cmd_get_displayable_str(cmd, NULL);
@@ -438,6 +451,172 @@ START_TEST (cmd_get_displayable_str_test) {
   ok = " bar baz";
   fail_if(res == NULL, "Expected string, got null");
   fail_unless(strcmp(res, ok) == 0, "Expected '%s', got '%s'", ok, res);
+
+  mark_point();
+  cmd = pr_cmd_alloc(p, 2, C_PASS, "foo");
+  res = pr_cmd_get_displayable_str(cmd, &len);
+  ok = "PASS (hidden)";
+  fail_unless(res != NULL, "Expected displayable string, got null");
+  fail_unless(len == 13, "Expected len 13, got %lu", (unsigned long) len);
+  fail_unless(strcmp(res, ok) == 0, "Expected '%s', got '%s'", ok, res);
+
+  mark_point();
+  cmd = pr_cmd_alloc(p, 2, C_ADAT, "bar baz quxx");
+  res = pr_cmd_get_displayable_str(cmd, &len);
+  ok = "ADAT (hidden)";
+  fail_unless(res != NULL, "Expected displayable string, got null");
+  fail_unless(len == 13, "Expected len 13, got %lu", (unsigned long) len);
+  fail_unless(strcmp(res, ok) == 0, "Expected '%s', got '%s'", ok, res);
+}
+END_TEST
+
+START_TEST (cmd_get_errno_test) {
+  int res, *xerrno = NULL;
+  cmd_rec *cmd = NULL;
+
+  res = pr_cmd_get_errno(NULL);
+  fail_unless(res == -1, "Failed to handle null cmd_rec");
+  fail_unless(errno == EINVAL, "Failed to set errno to EINVAL");
+
+  cmd = pr_cmd_alloc(p, 1, "foo");
+  res = pr_cmd_get_errno(cmd);
+  fail_unless(res == 0, "Expected errno 0, got %d", res);
+
+  (void) pr_table_remove(cmd->notes, "errno", NULL);
+  res = pr_cmd_get_errno(cmd);
+  fail_unless(res < 0, "Failed to handle missing 'errno' note");
+  fail_unless(errno == ENOENT, "Expected ENOENT (%d), got %s (%d)", ENOENT,
+    strerror(errno), errno);
+
+  xerrno = pcalloc(cmd->pool, sizeof(int));
+  (void) pr_table_add(cmd->notes, "errno", xerrno, sizeof(int));
+
+  res = pr_cmd_set_errno(NULL, ENOENT);
+  fail_unless(res < 0, "Failed to handle null arguments");
+  fail_unless(errno == EINVAL, "Expected EINVAL (%d), got %s (%d)", EINVAL,
+    strerror(errno), errno);
+
+  res = pr_cmd_set_errno(cmd, ENOENT);
+  fail_unless(res == 0, "Failed to stash errno ENOENT: %s", strerror(errno));
+
+  res = pr_cmd_get_errno(cmd);
+  fail_unless(res == ENOENT, "Expected errno ENOENT, got %s (%d)",
+    strerror(res), res);
+}
+END_TEST
+
+START_TEST (cmd_set_name_test) {
+  int res;
+  cmd_rec *cmd;
+  const char *name;
+
+  res = pr_cmd_set_name(NULL, NULL);
+  fail_unless(res < 0, "Failed to handle null arguments");
+  fail_unless(errno == EINVAL, "Expected EINVAL (%d), got %s (%d)", EINVAL,
+    strerror(errno), errno);
+
+  cmd = pr_cmd_alloc(p, 1, "foo");
+  res = pr_cmd_set_name(cmd, NULL);
+  fail_unless(res < 0, "Failed to handle null name");
+  fail_unless(errno == EINVAL, "Expected EINVAL (%d), got %s (%d)", EINVAL,
+    strerror(errno), errno);
+
+  name = "bar";
+  res = pr_cmd_set_name(cmd, name);
+  fail_unless(res == 0, "Failed to command name to '%s': %s", name,
+    strerror(errno));
+}
+END_TEST
+
+START_TEST (cmd_is_http_test) {
+  int res;
+  cmd_rec *cmd;
+
+  res = pr_cmd_is_http(NULL);
+  fail_unless(res < 0, "Failed to handle null arguments");
+  fail_unless(errno == EINVAL, "Expected EINVAL (%d), got %s (%d)", EINVAL,
+    strerror(errno), errno);
+
+  mark_point();
+  cmd = pr_cmd_alloc(p, 1, C_SYST);
+  cmd->argv[0] = NULL;
+  res = pr_cmd_is_http(cmd);
+  fail_unless(res < 0, "Failed to handle null name");
+  fail_unless(errno == EINVAL, "Expected EINVAL (%d), got %s (%d)", EINVAL,
+    strerror(errno), errno);
+
+  mark_point();
+  cmd->argv[0] = C_SYST;
+  res = pr_cmd_is_http(cmd);
+  fail_unless(res == FALSE, "Expected FALSE (%d), got %d", FALSE, res);
+
+  mark_point();
+  cmd = pr_cmd_alloc(p, 1, "GET");
+  res = pr_cmd_is_http(cmd);
+  fail_unless(res == TRUE, "Expected TRUE (%d), got %d", TRUE, res);
+}
+END_TEST
+
+START_TEST (cmd_is_smtp_test) {
+  int res;
+  cmd_rec *cmd;
+
+  res = pr_cmd_is_smtp(NULL);
+  fail_unless(res < 0, "Failed to handle null arguments");
+  fail_unless(errno == EINVAL, "Expected EINVAL (%d), got %s (%d)", EINVAL,
+    strerror(errno), errno);
+
+  mark_point();
+  cmd = pr_cmd_alloc(p, 1, C_SYST);
+  cmd->argv[0] = NULL;
+  res = pr_cmd_is_smtp(cmd);
+  fail_unless(res < 0, "Failed to handle null name");
+  fail_unless(errno == EINVAL, "Expected EINVAL (%d), got %s (%d)", EINVAL,
+    strerror(errno), errno);
+
+  mark_point();
+  cmd->argv[0] = C_SYST;
+  res = pr_cmd_is_smtp(cmd);
+  fail_unless(res == FALSE, "Expected FALSE (%d), got %d", FALSE, res);
+
+  mark_point();
+  cmd = pr_cmd_alloc(p, 1, "RCPT");
+  res = pr_cmd_is_smtp(cmd);
+  fail_unless(res == TRUE, "Expected TRUE (%d), got %d", TRUE, res);
+}
+END_TEST
+
+START_TEST (cmd_is_ssh2_test) {
+  int res;
+  cmd_rec *cmd;
+
+  res = pr_cmd_is_ssh2(NULL);
+  fail_unless(res < 0, "Failed to handle null arguments");
+  fail_unless(errno == EINVAL, "Expected EINVAL (%d), got %s (%d)", EINVAL,
+    strerror(errno), errno);
+
+  mark_point();
+  cmd = pr_cmd_alloc(p, 1, C_SYST);
+  cmd->argv[0] = NULL;
+  res = pr_cmd_is_ssh2(cmd);
+  fail_unless(res < 0, "Failed to handle null name");
+  fail_unless(errno == EINVAL, "Expected EINVAL (%d), got %s (%d)", EINVAL,
+    strerror(errno), errno);
+
+  mark_point();
+  cmd->argv[0] = C_SYST;
+  res = pr_cmd_is_ssh2(cmd);
+  fail_unless(res == FALSE, "Expected FALSE (%d), got %d", FALSE, res);
+
+  mark_point();
+  cmd = pr_cmd_alloc(p, 1, "SSH-2.0-OpenSSH_5.6p1");
+  res = pr_cmd_is_ssh2(cmd);
+  fail_unless(res == TRUE, "Expected TRUE (%d), got %d", TRUE, res);
+
+  mark_point();
+  cmd = pr_cmd_alloc(p, 1, "SSH-1.99-JSCH");
+  res = pr_cmd_is_ssh2(cmd);
+  fail_unless(res == TRUE, "Expected TRUE (%d), got %d", TRUE, res);
 }
 END_TEST
 
@@ -448,7 +627,6 @@ Suite *tests_get_cmd_suite(void) {
   suite = suite_create("cmd");
 
   testcase = tcase_create("base");
-
   tcase_add_checked_fixture(testcase, set_up, tear_down);
 
   tcase_add_test(testcase, cmd_alloc_test);
@@ -456,8 +634,12 @@ Suite *tests_get_cmd_suite(void) {
   tcase_add_test(testcase, cmd_cmp_test);
   tcase_add_test(testcase, cmd_strcmp_test);
   tcase_add_test(testcase, cmd_get_displayable_str_test);
+  tcase_add_test(testcase, cmd_get_errno_test);
+  tcase_add_test(testcase, cmd_set_name_test);
+  tcase_add_test(testcase, cmd_is_http_test);
+  tcase_add_test(testcase, cmd_is_smtp_test);
+  tcase_add_test(testcase, cmd_is_ssh2_test);
 
   suite_add_tcase(suite, testcase);
-
   return suite;
 }
