@@ -10,9 +10,9 @@ my $conn_ex;
 
 sub new {
   my $class = shift;
-  my ($addr, $port, $use_port, $timeout) = @_;
+  my ($addr, $port, $use_port, $conn_timeout, $cmd_timeout) = @_;
   $use_port = 0 unless defined($use_port);
-  $timeout = 10 unless defined($timeout);
+  $conn_timeout = 2 unless defined($conn_timeout);
  
   my $ftp;
 
@@ -24,7 +24,7 @@ sub new {
   # given timeout.
 
   my %opts = (
-      Port => $port,
+    Port => $port,
   );
 
   if ($use_port) {
@@ -38,9 +38,13 @@ sub new {
     $opts{Debug} = 10;
   }
 
+  if (defined($cmd_timeout)) {
+    $opts{Timeout} = $cmd_timeout;
+  }
+
   while (1) {
-    if (time() - $now > $timeout) {
-      croak("Unable to connect to $addr:$port: Timed out after $timeout secs");
+    if (time() - $now > $conn_timeout) {
+      croak("Unable to connect to $addr:$port: Timed out after $conn_timeout secs");
     }
 
     $ftp = Net::FTP->new($addr, %opts);
@@ -53,6 +57,16 @@ sub new {
     sleep(1);
   }
 
+  # Naughtily invade the Net::FTP internals; it makes for less confusion
+  # when writing the unit tests.
+  if (exists($ENV{FTP_FIREWALL})) {
+    ${*$ftp}{net_ftp_firewall} = $ENV{FTP_FIREWALL};
+  }
+
+  if (exists($ENV{FTP_FIREWALL_TYPE})) {
+    ${*$ftp}{net_ftp_firewall_type} = $ENV{FTP_FIREWALL_TYPE};
+  }
+
   my $self = {
     addr => $addr,
     ftp => $ftp,
@@ -72,8 +86,12 @@ sub response_code {
 
 sub response_msg {
   my $self = shift;
-  my $index = shift;
-  $index = 1 unless defined($index);
+  my $req_index = shift;
+
+  my $index = 1;
+  if (defined($req_index)) {
+    $index = $req_index;
+  }
 
   if (defined($self->{mesg})) {
     my $msg = $self->{mesg};
@@ -83,12 +101,27 @@ sub response_msg {
   }
 
   my @msgs = $self->{ftp}->message;
-  if (scalar(@msgs) > 1) {
+  my $nmsgs = scalar(@msgs);
+  if ($nmsgs > 1) {
+    if ($index > ($nmsgs - 1)) {
+      return undef;
+    }
+
     chomp($msgs[$index]);
     return $msgs[$index];
+
+  } else {
+    if (defined($req_index)) {
+      if ($index > 0) {
+        return undef;
+      }
+    }
+  }
+
+  if (defined($msgs[0])) {
+    chomp($msgs[0]);
   }
 
-  chomp($msgs[0]);
   return $msgs[0];
 }
 
@@ -151,7 +184,22 @@ sub login {
   $SIG{ALRM} = \&login_alarm;
   alarm($login_timeout);
 
+  # Work around some (strange? broken?) ness in Net::FTP's handling of
+  # the destination server in the login() method for "firewalls"
+  # (i.e. proxying).
+
+  my $ftp = $self->{ftp};
+  my $net_ftp_host = ${*$ftp}{net_ftp_host};
+
+  if (exists($ENV{FTP_FIREWALL})) {
+    ${*$ftp}{net_ftp_host} = $ENV{FTP_FIREWALL};
+  }
+
   unless ($self->{ftp}->login($user, $pass)) {
+    if (exists($ENV{FTP_FIREWALL})) {
+      ${*$ftp}{net_ftp_host} = $net_ftp_host;
+    }
+
     alarm(0);
     $SIG{ALRM} = 'DEFAULT';
 
@@ -1582,7 +1630,7 @@ sub stat {
   }
 }
 
-# From the FTP HOST command RFCXXXX (currently in Draft form)
+# From the FTP HOST command RFC 7151
 sub host {
   my $self = shift;
   my $host = shift;
@@ -1609,4 +1657,47 @@ sub host {
   }
 }
 
+sub clnt {
+  my $self = shift;
+  my $info = shift;
+  $info = 'ProFTPD::TestSuite::FTP' unless defined($info);
+  my $code;
+
+  $code = $self->{ftp}->quot('CLNT', $info);
+  unless ($code) {
+    croak("CLNT command failed: " .  $self->{ftp}->code . ' ' .
+      $self->response_msg());
+  }
+
+  if ($code == 4 || $code == 5) {
+    croak("CLNT command failed: " .  $self->{ftp}->code . ' ' .
+      $self->response_msg());
+  }
+
+  my $msg = $self->response_msg();
+  if (wantarray()) {
+    return ($self->{ftp}->code, $msg);
+
+  } else {
+    return $msg;
+  }
+}
+
+sub abort {
+  my $self = shift;
+
+  unless ($self->{ftp}->abort()) {
+    croak("ABOR command failed: " .  $self->{ftp}->code . ' ' .
+      $self->response_msg());
+  }
+
+  my $msg = $self->response_msg();
+  if (wantarray()) {
+    return ($self->{ftp}->code, $msg);
+
+  } else {
+    return $msg;
+  }
+}
+
 1;
